import copy
from json import JSONDecodeError
from typing import Optional, Dict

import json
from datetime import datetime
from io import StringIO

from ytrss.core.entity.movie import Movie, MovieError
from ytrss.core.helpers.typing import Url
from ytrss.plugins.youtube_dl.wrapper import youtube_main_wrapper


class InvalidDataMovieError(MovieError):
    """ Invalid parameter movie error

    This exception is invoked when the movie data is invalid
    """


class YouTubeMovie(Movie):
    """ A YouTube implementation of movie """

    def __init__(self, url: Url) -> None:
        self._url: Url = url
        self._date: Optional[datetime] = None
        self._json_data = self._get_movie_data()

    def _get_movie_data(self) -> Dict[str, str]:
        _, json_output, self._error = youtube_main_wrapper('--dump-json', self._url)
        try:
            output = json.load(StringIO(json_output))
            if isinstance(output, dict):
                return output
            raise InvalidDataMovieError(f"output element is not dictionary {type(output)}")
        except JSONDecodeError as exc:
            raise InvalidDataMovieError(f"Invalid address type: [{self._url}] -> {self._error}") from exc

    @property
    def url(self) -> Url:
        return self._url

    @property
    def identity(self) -> str:
        extractor: str = self.__get_youtube_data("extractor")
        code: str = self.__get_youtube_data("display_id")
        return f"ytdl:{extractor}:{code}"

    def __get_youtube_data(self, key: str) -> str:
        return str(self._json_data.get(key, ""))

    @property
    def title(self) -> str:
        return self.__get_youtube_data("title")

    @property
    def author(self) -> str:
        return self.__get_youtube_data('uploader')

    @property
    def description(self) -> str:
        return self.__get_youtube_data("description")

    @property
    def date(self) -> datetime:
        try:
            return datetime.strptime(self.__get_youtube_data("upload_date"), "%Y%m%d")
        except ValueError:
            return datetime.now()

    @property
    def img_url(self) -> Optional[Url]:
        img = self.__get_youtube_data("thumbnail")
        return Url(img) if img != "" else None

    @property
    def is_ready(self) -> bool:
        is_live = "false"
        if self._error is None:
            is_live = self.__get_youtube_data("is_live")

        return ("Premieres in" not in self._error if self._error is not None else True) and is_live.lower() != "true"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Movie):
            return self.identity == other.identity
        if other == "":
            return False
        tmp_other = copy.deepcopy(other)
        return tmp_other == self

    def __str__(self) -> str:
        return self.url



import abc
from dataclasses import dataclass

from ytrss.core.entity.movie import Movie


@dataclass
class WebhookType(metaclass=abc.ABCMeta):
    """ An abstract class representing a webhook """
    name: str
    data: dict[str, str]


class StartDaemonWebhook(WebhookType):
    """ A webhook class representing a start daemon """

    def __init__(self) -> None:
        super().__init__('start-daemon', {})


class FoundNewMovieWebhook(WebhookType):
    """ A webhook class representing a found new movie """
    def __init__(self, movie: Movie, destination: str) -> None:
        super().__init__('found-new-movie', {
            'title': movie.title,
            'url': movie.url,
            'id': movie.identity,
            'desc': movie.description,
            'destination': destination
        })


class StartDownloadingMovieWebhook(WebhookType):
    """ A webhook class representing a start downloading movie """
    def __init__(self, movie: Movie) -> None:
        super().__init__('start-downloading-movie', {
            'title': movie.title,
            'url': movie.url,
            'id': movie.identity,
            'desc': movie.description
        })


class SuccessDownloadMovieWebhook(WebhookType):
    """ A webhook class representing a download movie """
    def __init__(self, movie: Movie) -> None:
        super().__init__('success-download-movie', {
            'title': movie.title,
            'url': movie.url,
            'id': movie.identity,
            'desc': movie.description
        })


class FailedDownloadMovieWebhook(WebhookType):
    """ A webhook class representing a failed downloaded movie """
    def __init__(self, movie: Movie, cause: Exception) -> None:
        super().__init__('failed-download-movie', {
            'title': movie.title,
            'url': movie.url,
            'id': movie.identity,
            'desc': movie.description,
            'cause': str(cause)
        })



import sys
from io import StringIO, TextIOBase
from typing import Union
import yt_dlp
from ytrss.core.helpers.logging import logger


class _Tee(TextIOBase):
    """
    A TextIO that display elements on screen and save to String
    """
    def __init__(self, is_error: bool = False) -> None:
        self._string = StringIO()
        self._is_error = is_error

    def write(self, obj: str) -> int:
        """
        Write object to stream
        """
        if self._is_error:
            logger.error(obj)
        else:
            logger.info(obj)
        return self._string.write(obj)

    def flush(self) -> None:
        """
        Flush string io
        """
        self._string.flush()

    def getvalue(self) -> str:
        """
        Get string value.
        """
        return self._string.getvalue()


def youtube_main_wrapper(*args: str, show_output: bool = False) -> tuple[int, str, str]:
    """
    A wrapper to yt_dlp function with output values
    """
    old_stdout = sys.stdout
    old_stderr = sys.stderr

    tmp_stdout: TextIOBase
    tmp_stderr: TextIOBase

    if show_output:
        sys.stdout = tmp_stdout = _Tee()
        sys.stderr = tmp_stderr = _Tee(is_error=True)
    else:
        sys.stdout = tmp_stdout = StringIO()
        sys.stderr = tmp_stderr = StringIO()

    try:
        logger.debug("Invoke command: yt-dlp %s", " ".join(args))

        yt_dlp.main(list(args))
        status: Union[str, int] = 0
    except SystemExit as ex:
        if ex.code is None:
            status = 0
        else:
            status = ex.code  # pylint: disable=E0012,R0204

    sys.stdout.flush()
    sys.stderr.flush()

    if isinstance(status, str):
        status = 0
    sys.stdout = old_stdout
    sys.stderr = old_stderr

    return status, tmp_stdout.getvalue(), tmp_stderr.getvalue()



from argparse import Namespace

from ytrss.commands import BaseCommand
from ytrss.core.algoritms.generate import generate


class GenerateCommand(BaseCommand):
    """
    Generate elements in destinations.
    """
    def __init__(self) -> None:
        BaseCommand.__init__(self, "generate")

    async def run(self, options: Namespace) -> int:
        await generate(self.manager_service)
        return 0



from datetime import datetime
import abc

from enum import Enum

from typing import Iterable, Tuple, Optional, Union

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.core.entity.downloaded_movie import DownloadedMovie
from ytrss.core.entity.movie import Movie

MovieSelect = Union[Movie, DownloadedMovie, str]


def get_movie_id(movie: MovieSelect) -> str:
    """ Get movie id. """
    if isinstance(movie, DownloadedMovie):
        return movie.identity
    if isinstance(movie, Movie):
        return movie.identity
    return movie


class DatabaseStatus(Enum):
    """
    Database status types
    """
    WAIT = "wait"
    PROGRESS = "progress"
    DONE = "done"
    ERROR = "error"
    DELETED = "deleted"


class Database(metaclass=abc.ABCMeta):
    """
    Abstract database class
    """

    @abc.abstractmethod
    def is_new(self, movie: MovieSelect, destination: DestinationId) -> bool:
        """ Check if movie is ready to download. """

    @abc.abstractmethod
    def change_type(self, movie: MovieSelect, db_type: DatabaseStatus) -> None:
        """ Change type of movie task. """

    @abc.abstractmethod
    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        """ Get list of movies to download. """

    @abc.abstractmethod
    def queue_mp3(self, movie: Movie, destination: DestinationId) -> bool:
        """ Add movie task to queue. """

    @abc.abstractmethod
    def get_created_data(self, movie: MovieSelect) -> Optional[datetime]:
        """ Get date of movie create. """



from dataclasses import dataclass
from datetime import datetime
from typing import List, Iterator
from ytrss.core.entity.downloaded_movie import DownloadedMovie
from ytrss.core.helpers.logging import logger
from ytrss.core.managers.manager_service import ManagerService, default_manager_service
from ytrss.database.database import DatabaseStatus


@dataclass
class _MovieWithData:
    def __init__(self, movie: DownloadedMovie, date: datetime) -> None:
        self.movie = movie
        self.date = date


def filter_limits(
        manager_service: ManagerService,
        movies: List[DownloadedMovie],
        limit: int = 20
) -> Iterator[DownloadedMovie]:
    """ Return a list of movies that are pass to limits """

    movie_list: List[_MovieWithData] = []

    for movie in movies:
        create_date = manager_service.database.get_created_data(movie)
        if create_date is None:
            yield movie
        else:
            movie_list.append(_MovieWithData(movie, create_date))

    sorted_movies = sorted(movie_list, key=lambda movie_data: movie_data.date)
    yield from map(lambda movie_data: movie_data.movie, sorted_movies[:-limit])


async def clean_tasks(
    manager_service: ManagerService = default_manager_service()
) -> None:
    """ Clean tasks """

    for destination in manager_service.destination_manager.destinations:
        deleted_movies = destination.saved_movies
        deleted_movies = filter_limits(manager_service, list(deleted_movies), limit = destination.info.limit)
        print("Start deleting")
        for movie in deleted_movies:
            logger.info("Movie deleted: [%s] %s",
                        movie.url,
                        movie.title)
            destination.delete(movie, manager_service.templates_manager)
            manager_service.database.change_type(movie, DatabaseStatus.DELETED)



import os
import time
from datetime import datetime
from argparse import Namespace

from locks import Mutex

from ytrss.commands.configuration import ConfigurationCommand
from ytrss.commands.version import VersionCommand
from ytrss.commands import BaseCommand
from ytrss.core.algoritms.clean import clean_tasks
from ytrss.core.algoritms.download import download_all_movies
from ytrss.core.algoritms.finder import prepare_urls
from ytrss.core.entity.webhooks import StartDaemonWebhook
from ytrss.core.helpers.logging import logger


class DaemonCommand(BaseCommand):
    """
    Run downloading in loop mode
    """
    def __init__(self) -> None:
        BaseCommand.__init__(self, "daemon")

    async def run(self, options: Namespace) -> int:

        self.manager_service.webhook_manager.invoke_hook(StartDaemonWebhook())

        logger.info("""
 -------------------------------------------------

             /$$
            | $$
 /$$   /$$ /$$$$$$    /$$$$$$   /$$$$$$$ /$$$$$$$
| $$  | $$|_  $$_/   /$$__  $$ /$$_____//$$_____/
| $$  | $$  | $$    | $$  \\__/|  $$$$$$|  $$$$$$
| $$  | $$  | $$ /$$| $$       \\____  $$\\____  $$
|  $$$$$$$  |  $$$$/| $$       /$$$$$$$//$$$$$$$/
 \\____  $$   \\___/  |__/      |_______/|_______/
 /$$  | $$
|  $$$$$$/
 \\______/

 -------------------------------------------------
        """)

        logger.info("")
        await VersionCommand().run(options)
        await ConfigurationCommand().run(options)
        logger.info("")
        while True:
            os.makedirs('/tmp/ytrss', exist_ok=True)
            try:
                with Mutex('/tmp/ytrss/ytrss.lock'):
                    logger.info("Start process: %s", datetime.now())
                    prepare_urls()
                    await download_all_movies()
                    await clean_tasks()
                    logger.info("Stop process: %s", datetime.now())
            except BlockingIOError:
                logger.info("Program is already started")
                return 1
            except Exception as ex:  # pylint: disable=W0703
                logger.error("Unexpected daemon problem: %s", str(ex))

            time.sleep(3600)



import requests

from ytrss.configuration.entity.configuration_data import YtrssConfiguration
from ytrss.core.entity.webhooks import WebhookType
from ytrss.core.helpers.logging import logger


class WebhookManager:
    """
    Webhook Manager

    That object is responsible for managing webhooks. It can store and invoke webhook when its necessary. All
    webhooks should be defined in WebhookType, it has different name and set of arguments.
    """

    def __init__(self, configuration: YtrssConfiguration) -> None:
        self._configuration = configuration

    def invoke_hook(self, webhook: WebhookType) -> None:
        """ Invoke a webhook """

        logger.debug("[Webhook] Invoke Webhook: %s", webhook.name)
        for webhook_url in self._configuration.get_webhook(webhook.name):
            webhook_url_formated = webhook_url.format(**webhook.data)
            try:
                response = requests.get(webhook_url_formated, timeout=10)

                if response.status_code == 200:
                    logger.debug("[Webhook] Invoke url webhook: %s, %s", webhook.name, webhook_url_formated)
                else:
                    logger.error(
                        "[Webhook] Address url (%s) has %s output code",
                        webhook_url_formated,
                        response.status_code
                    )
            except requests.exceptions.RequestException as e:
                logger.error("[Webhook] Problem with request (%s): %s", webhook_url_formated, str(e))



from argparse import Namespace

from ytrss.commands import BaseCommand
from ytrss.core.helpers.logging import logger


class ConfigurationCommand(BaseCommand):
    """
    Print configuration
    """

    def __init__(self) -> None:
        BaseCommand.__init__(self, "configuration")

    async def run(self, options: Namespace) -> int:
        logger.info("subscriptions:")
        for source in self.manager_service.configuration.sources:
            logger.info("\t%s [%s]%s (%s)",
                        source.name,
                        source.url,
                        " [disabled]" if not source.enable else "",
                        source.destination)

        logger.info("\ndestinations:")
        for destination in self.manager_service.destination_manager.destinations:
            logger.info("\t[%s] => %s", destination.identity, destination.info.title)
            logger.info("\t\toutput_path => %s", destination.info.output_path)
            logger.info("\t\turl => %s", destination.info.url_prefix)

        logger.info("\nplugins:")
        for plugin in self.manager_service.plugin_manager.plugins:
            logger.info("\t[%s] => %s", plugin.plugin_name, plugin.plugin_description)

        logger.info("\nwebhooks:")
        webhooks = self.manager_service.configuration.webhooks
        for webhook in webhooks:
            logger.info("\t[%s] =>", webhook)
            for webhook_url in self.manager_service.configuration.get_webhook(webhook):
                logger.info("\t  - %s", webhook_url)
        if not webhooks:
            logger.info("\t No webhook found")
        return 0



import os
import shutil
from typing import Iterator, Sequence

from ytrss.core.entity.destination import Destination, DestinationError

from ytrss.core.entity.downloaded_movie import DownloadedMovie, MovieFileError

from ytrss.configuration.entity.destination_info import DestinationId, DestinationInfo
from ytrss.core.helpers.logging import logger
from ytrss.core.helpers.typing import Path
from ytrss.core.managers.templates_manager import TemplatesManager


class DefaultDestination(Destination):
    """
    A destination copy files to destination folder.
    """

    def __init__(self, info: DestinationInfo) -> None:
        # pylint: disable=C0123
        if type(self) == DefaultDestination and info.type != 'default' and info.type != 'copy':
            raise DestinationError()
        self.__info = info

    @property
    def identity(self) -> DestinationId:
        return self.__info.identity

    @property
    def info(self) -> DestinationInfo:
        return self.__info

    @property
    def saved_movies(self) -> Iterator[DownloadedMovie]:
        if self.info.output_path is None:
            logger.error("Destination %s output_path is None", self.identity)
            return
        try:
            for filename in os.listdir(self.info.output_path):
                if filename.endswith(".json"):
                    try:
                        yield DownloadedMovie.from_file(self.info.output_path, filename)
                    except MovieFileError as ex:
                        logger.info(ex)
        except FileNotFoundError as ex:
            logger.warning("Destination %s directory not exists: %s", self.identity, ex.filename)

    def save(self, files: Sequence[Path], templates_manager: TemplatesManager) -> None:
        if self.info.output_path is None:
            raise KeyError
        os.makedirs(self.info.output_path, exist_ok=True)
        for file in files:
            try:
                shutil.move(file, self.info.output_path)
            except shutil.Error as ex:
                logger.error(ex)
        self.on_finish(templates_manager)

    def delete(self, movie: DownloadedMovie, templates_manager: TemplatesManager) -> None:
        for file_path in movie.data_paths:
            try:
                os.remove(file_path)
                logger.info("\tdelete file: %s from %s", file_path, movie.identity)
            except Exception as ex:  # pylint: disable=W0718
                logger.error("Cannot remove %s: %s", file_path, str(ex))
        self.on_finish(templates_manager)



class ConfigurationError(Exception):
    """ Configuration error """


class ConfigurationFileNotExistsError(ConfigurationError):
    """ Configuration file not exists error

    The error that raised when the configuration file is not existed.
    """



from argparse import Namespace

from ytrss.commands import BaseCommand
from ytrss.core.algoritms.clean import clean_tasks


class CleanCommand(BaseCommand):
    """
    Clean all unnecessary files.
    """
    def __init__(self) -> None:
        BaseCommand.__init__(self, "clean")

    async def run(self, options: Namespace) -> int:
        await clean_tasks()
        return 0



from typing import Iterable, Tuple

from ytrss.configuration.entity.destination_info import DestinationId
from ytrss.configuration.entity.source import Source
from ytrss.core.entity.movie import Movie
from ytrss.core.entity.source_downloader import SourceDownloader, SourceDownloaderError
from ytrss.core.helpers.typing import Url
from ytrss.plugins.youtube_dl.movie import YouTubeMovie
from ytrss.plugins.youtube_dl.wrapper import youtube_main_wrapper


class YouTubeDlSourceDownloader(SourceDownloader):
    """ YouTube Source Downloader

    This object put a list of movies from source if its support by `yt-dlp`
    """

    def __init__(self, source: Source) -> None:
        self.url = source.url
        self.destination = source.destination
        status, output, error = youtube_main_wrapper(
            '--flat-playlist',
            '--playlist-end',
            '20', '--print',
            'url',
            self.url
        )
        self.output = output
        if status != 0:
            raise SourceDownloaderError(f'Invalid url source (self.url): {error}')

    @property
    def movies(self) -> Iterable[Tuple[Movie, DestinationId]]:
        for url in self.output.splitlines():
            try:
                yield YouTubeMovie(Url(url)), self.destination
            except Exception:  # pylint: disable=W0703
                pass



from ytrss.core.managers.manager_service import ManagerService, default_manager_service


async def generate(manager_service: ManagerService = default_manager_service()) -> None:
    """ Generate destination files. """
    for destination in manager_service.destination_manager.destinations:
        destination.on_finish(manager_service.templates_manager)



import os

from ytrss.configuration.entity.configuration_data import YtrssConfiguration
from ytrss.core.entity.downloaded_movie import DownloadedMovie
from ytrss.core.entity.movie import Movie
from ytrss.core.helpers.exceptions import DownloadMovieError

from ytrss.core.helpers.files import cwd
from ytrss.core.helpers.logging import logger
from ytrss.core.helpers.typing import Path
from ytrss.plugins.youtube_dl.movie import YouTubeMovie
from ytrss.plugins.youtube_dl.wrapper import youtube_main_wrapper


class YouTubeDownloader:
    """
    YouTube downloader

    Downloader that use `yt_dlp <https://rg3.github.io/youtube-dl/>`_ implementation. It downloads
    a movie file from any url that can be use by `yt_dlp` service.
    """

    def __init__(self, movie: Movie, configuration: YtrssConfiguration) -> None:
        if not isinstance(movie, YouTubeMovie):
            raise DownloadMovieError("YouTubeMovie instance not exists")
        self.configuration = configuration
        self._movie = movie

    def download(self) -> DownloadedMovie:
        """ Download movie using YouTube_dl tool """
        try:
            os.makedirs(self.configuration.cache_path)
        except OSError:
            pass

        with cwd(self.configuration.cache_path):
            logger.info("Downloading movie: [%s] %s", self._movie.url, self._movie.title)
            status, _, _ = youtube_main_wrapper(
                *self.configuration.args,
                '-o',
                self._movie.identity,
                self._movie.url,
                show_output=True
            )

            if status != 0:
                raise DownloadMovieError(f"yt_dlp raise with state: {status}")

            for extension in ['mp3', 'mp4', 'webm', 'mkv', 'ogg']:
                full_file_name = Path(f"{self._movie.identity}.{extension}")
                if os.path.isfile(full_file_name):
                    return DownloadedMovie.create_movie_desc(self.configuration.cache_path, self._movie,
                                                             [full_file_name])
        raise DownloadMovieError(f"File {self._movie.identity} not downloaded")
